      subroutine sasas(ier,numor,inst,nomexp,ddate,nt,text,np,
     1par,nd,idon)
c Special cache version for splots and pdqheads (xplots)
c
c     reading standard SANS ascii files, or D16 data files...
c     which may be compressed.. (*.Z files)
c     ier             integer      out  0 ok, else -1
c     numor           integer      in   1<=numor<=999999
c     inst            character*4  out  ints field in file
c     nomexp          character*10 out  nomexp in file
c     ddate           character*20 out  data and time of recording
c     nt              integer      out  number of text (a1)
c     text            character*1  out  text array
c     np              integer      out  number of real parameters
c     par             real         out  real parameters
c     nd              integer      out  number of integers in spectrum
c     idon            integer      out  spectrum
c environment variables
c  SAS_INSTRUMENT     instrument name 4 characters
c  SAS_DATA_PATH      directory for raw data
c  SAS_DEBUG    flag to print debugging information
c  SAS_DECOMPRESS zcat  "zcat" is program name of decompresion tool
c     R. E. Ghosh  March 1996   ron@ill.fr
c  SAS_CACHE     see note October 2010
c  the commons:
c  /maxarr/ ntitl,npar,nparx,ndx  512,128,20,16384 respectively
c           protect programs from excessive input. These stored 
c           defaults may be overwritten by calling programs
c  /sasasc/ nunit  file input unit is set to 31 by default
c
c  /liminp/ ininp  is set to -1 by default. If positive then only
c           that number of data fields are read-in (now there are 5) 
c     History of overflows - continued
C*****  SANS Data Storage - Program MTD  (R. Ghosh 1981-89)
C
C       and now a bit of history:
C       IB(249) defined in documentation should be 1 for 16bit, 3 for 32bit
C       data storage. (ILL-82GH05G, 89GH03T)
C       IB(253) used to be used for TOF instruments to indicate whether data 
C       were to beconverted to 32 bit before storage on DECsystem10. 
C       When 0 no action was to be taken 
C       (otherwise this indicated the position of the
C       letter K in the internal ASCII format: letter this had to be modified
C       to L and IB(253) reset to 0 signalling transormation to 32bit data.) 
C       On the SANS instruments IB(249) has always been set to 0
C       and IB(253)=200 to show that data are 16 bit, because data were
C       subject to special overflow tretment on the DEC-10.
C       No change was introduced when archiving on the central VAX started.
C       All data on VAX, restored from DEC-10 tapes are restored as 32bits
C       and correctly IB(249) is set to 3, and IB(253) is set to 0,
C       following above definitions.
C       From today (December 7th 1989) the test for SANS data will be on 
C       IB(249). If this is set to 3 then 32bit data is assumed, otherwise
C       16 bits is the default. This accounts satisfactorily for all cases.
C       Perhaps one day the header block on the instruments will be corrected
C       to conform 100% with the documentation!
C
C       March 1995 it has happened! - ib(249) medpar(149) is 1 for 16 bit
C       acquisition
C
C       March 1996 adapted to read in either SANS (D11,D17,D22) or D16 data
C       September 1999
C       for D16  version 4 the sequence of counters difers from
C       version 3 and this has now been corrected.  Note that
C       the time for post 1998 data is now initially 1khz.
C       October 1999 will open compressed data files (*.Z) if
C       SAS_DECOMPRESS is set to the decompress program "zcat"
C       Modified March 2002 for new detector on D16, common/sasinx added
C       Modified May 2004 to include a date from d16
C        also first row and column for D16 reset to 0
C        nominal X0,Y0, SD given for D16
C        sample temperature constrained 0-4000K
C       Modified May 2008 to account for .Z or .gz files
C       Modified October 2010 (cropen) for using temporary cache (cashas)
C        or directory named in SAS_CACHE environment variable (no terminator)
C        SAS_CACHE can be managed by xencash tool
C
C       Modified November 2011 - added D16 paras - flag version mismatch 
      common/maxarr/ntitl,npar,nparx,ndx
      common/sasasc/nunit
      common/sasmed/medpar(156)
      common/sasinx/instx
      common/liminp/ininp
      common/devfil/fdev
      common/devfii/lf
      common/sdbg/ideb
      real par(128)
      integer idon(1)
      character*8 rtype
      character inst*4,instx*4,nomexp*10,ddate*20
      character fnum*6,fdev*66,errmsg*16,gdev*60,debug*20
      character*1 text(1)
c****** 
c
cd22:-  For reading 156I, 512A, 128F, 256I(overflows or skip), 16384I
cd16:-               80a, 31I, 50F, 1034I (typically)  
cd16-2002            80a, 31I, 50F, (4F 16384I)
cd16-2011            80a, 31I, 60F, (5F 102400I) (version5)
c******
      ivers=1
c***** modified for D16  September 1997 (REG), Nov 2011
      inst=' '
      ddate=' '
      nomexp=' '
      np=0
      nt=0
      nd=0
      do 500 i=1,npar
      par(i)=0.
500   continue
      do 501 i=1,ntitl
      text(i)=' '
501   continue
c***** local reals initialised
       indone=ininp
      if(indone.gt.5) indone=-1
c****** there are 5 data fields present in current data
      ier=-1
      call getenv('SAS_DATA_PATH',gdev)
      call getenv('SAS_INSTRUMENT',instx)
      call getenv('SAS_DEBUG',debug)
c***** get length
      call trimmd(gdev,lf)
      call toup(instx)
      if(instx(1:3).eq.'D16') ivers=10
c***** modified for D16  September 1997 (REG), November 2011
      ideb=index(debug,' ')
c***** debug is all blank if not set, hence ideb=1 for no debug
      if(ideb.ne.1) write(6,95) instx,gdev 
95    format(' SASAS - seeking data for instrument ',a4/
     1       '         in directory ',a)
      fnum=' '
      fdev=' '
      errmsg='numor'
      if(ideb.ne.1) write(6,97) errmsg
 97    format(' SASAS - debug reached test for ',a)
      if(numor.le.0.or.numor.gt.999999) go to 99
      write(fnum,1) numor
1     format(i6.6)
      errmsg='unit'
      if(nunit.le.0) go to 99
      fdev=gdev(1:lf)//fnum
c***** the device name must include the separator /,\ or :
      lf=lf+6
      errmsg='opening file'
      if(ideb.ne.1) write(6,97) errmsg
      call cropen(fdev,lf,nunit,ier)
      if(ier.ne.0) go to 99
c      open(unit=nunit,file=fdev,status='old',err=99)
      errmsg='record 1'
      read(nunit,2,err=99,end=99) rtype
2     format(a)
      if(rtype.ne.'RRRRRRRR') go to 99
      errmsg='numor'
      if(ideb.ne.1) write(6,97) errmsg
      read(nunit,3,err=99,end=99) numr,itxt,ivrs
3     format(10i8)
      errmsg='version'
      if(ideb.ne.1) write(6,97) errmsg
      if(ivers.lt.ivrs) go to 99
      if(itxt.gt.0) call skip(itxt)
      errmsg='header text'
      if(ideb.ne.1) write(6,97) errmsg
      read(nunit,2,err=99,end=99) rtype
      read(nunit,3,err=99,end=99) nchars,itxt
      if(itxt.gt.0) call skip(itxt)
      read(nunit,5,err=99,end=99) inst,nomexp,ddate
5     format(3a)
      errmsg='inst'
      if(ideb.ne.1) write(6,97) errmsg
      if(inst.ne.instx) go to 99
c***** treat SANS data and D16 data separately
      if(inst.eq.'D16') 
     1call indif(ier,ideb,indone,ddate,nt,text,np,par,nd,idon)
      if(inst.ne.'D16')
     1call insas(ier,ideb,indone,nt,text,np,par,nd,idon)
c***** check against delivering unbelievable sample temperatures.
      if(par(31).gt.4000..or.par(31).lt.0.) par(31)=0.
      return
99    write(6,98) errmsg,fdev(1:lf)
98    format(' SASAS: Error - invalid - ',a,/
     1' Currently reading : ',a)
      close(nunit)
      return
      end
      subroutine insas(ier,ideb,indone,nt,text,np,par,nd,idon)
c***** reads standard fields expecting:
c       156I, 512A, 128F, 256I(overflows or skip), (4096 or 16384)I:
c
      common/maxarr/ntitl,npar,nparx,ndx
      common/sasasc/nunit
      common/sasmed/medpar(156)
      common/liminp/ininp
      common/devfil/fdev
      common/devfii/lf
      character fdev*66,errmsg*16
      character*1 text(1)
      character*8 rtype
      real par(1)
      integer idon(1),iovs(256)
      errmsg='medpar array'
      if(ideb.ne.1) write(6,97) errmsg
97    format(' SASAS - debug reached test for ',a)
      read(nunit,2,err=99,end=99) rtype
      if(rtype.ne.'IIIIIIII') go to 99
      read(nunit,3,err=99,end=99) ni,itxt
      if(ni.ne.156) go to 99
      if(itxt.gt.0) call skip(itxt)
      read(nunit,3,err=99,end=99) medpar
3     format(10i8)
      indone=indone-1
      if(indone.eq.0) go to 20
      errmsg='text array'
      if(ideb.ne.1) write(6,97) errmsg
      read(nunit,2,err=99,end=99) rtype
2     format(a)
      if(rtype.ne.'AAAAAAAA') go to 99
      read(nunit,3,err=99,end=99) nt,itxt
      if(itxt.gt.0) call skip(itxt)
      errmsg='too much text'
      if(ideb.ne.1) write(6,97) errmsg
      if(nt.gt.ntitl) go to 99
      read(nunit,6,err=99,end=99) (text(i),i=1,nt)
6     format(80a1)
      indone=indone-1
      if(indone.eq.0) go to 20
      errmsg='real parameters'
      if(ideb.ne.1) write(6,97) errmsg
      read(nunit,2,err=99,end=99) rtype
      if(rtype.ne.'FFFFFFFF') go to 99
      read(nunit,3,err=99,end=99) np,itxt
      if(np.gt.npar) go to 99
      if(itxt.gt.0) call skip(itxt)
      read(nunit,7,err=99,end=99) (par(i),i=1,np)
7     format(5e16.8)
      indone=indone-1
      if(indone.eq.0) go to 20
c***** overflows to be added in - else skip to spectrum....
      if(medpar(149).eq.3) go to 10
c***** these are already 32 bit data
c***** no overflows are stored
      novs=-medpar(40)*2
c***** overflows stored as channel/count channel/count pairs
      if(novs.ge.0) go to 10
c***** read overflows for 16 bit spectra
      errmsg='overflows'
      if(ideb.ne.1) write(6,97) errmsg
      read(nunit,2,err=99,end=99) rtype
      if(rtype.ne.'IIIIIIII') go to 99
      read(nunit,3,err=99,end=99) nov,itxt
      if(itxt.gt.0) call skip(itxt)
      read(nunit,3,err=99,end=99) (iovs(i),i=1,nov)
c***** next field should be SSSS
10    continue
      indone=indone-1
      if(indone.eq.0) go to 20
c***** skip to SSSS
12    read(nunit,2,err=99,end=99) rtype
      if(rtype.ne.'SSSSSSSS') go to 12
      errmsg='spectrum header'
      if(ideb.ne.1) write(6,97) errmsg
      read(nunit,3,err=99,end=99) ispec,nrest,ntot,nrun,itxt
      if(itxt.gt.0) call skip(itxt)
      if(ispec.ne.1) go to 99
c***** should be one spectrum only
      errmsg='spectrum data'
      if(ideb.ne.1) write(6,97) errmsg
      read(nunit,2,err=99,end=99) rtype
      if(rtype.ne.'IIIIIIII') go to 99
      read(nunit,3,err=99,end=99) nd,itxt
      if(itxt.gt.0) call skip(itxt)
      errmsg='too many data'
      if(ideb.ne.1) write(6,97) errmsg
      if(nd.gt.ndx) go to 99
      errmsg='counts data'
      read(nunit,3,err=99,end=99) (idon(i),i=1,nd)
c***** add in overflows if any
      if(medpar(149).eq.3) go to 20
      if(novs.le.0) go to 20
      do 16 i=1,novs,2
      iad=iovs(i)+1
c***** address starts at 0
      if(iad.gt.0.and.iad.le.nd)
     1idon(iad)=idon(iad)+iovs(i+1)*65536
16    continue
20    close(nunit)
      ier=0
      return
99    write(6,98) errmsg,fdev(1:lf)
98    format(' SASAS: Error - invalid - ',a,/
     1' Currently reading : ',a)
      close(nunit)
      return
      end
      subroutine indif(ier,ideb,indone,ddate,nt,text,np,par,nd,idon)
c***** expecting standard input:
c      nvers=first integer value (in 31 integer field)
c           80a, 31I, 50F, 1034I (typically) for nvers=3
c           80a, 31I, 50F, 5F, 1024I  for nvers=4 (annotated)
c           80a, 31I, 60F, 5f, 102400I for nvers=5
      common/maxarr/ntitl,npar,nparx,ndx
      common/sasasc/nunit
      common/sasmed/medpar(156)
      common/liminp/ininp
      common/devfil/fdev
      common/devfii/lf
      character*12 D16_SD_M
      character fdev*66,errmsg*16
      character*1 text(1)
      character*8 rtype
      character*(*) ddate
      real par(*),parr(200),parrs(10)
      integer idon(1),iflag
      data iflag/0/
      save iflag
      nt=0
      ni=0
      np=0
      nps=0
      errmsg='text array'
      if(ideb.ne.1) write(6,97) errmsg
 97    format(' SASAS - debug reached test for ',a)
      read(nunit,2,err=99,end=99) rtype
      if(rtype.ne.'AAAAAAAA') go to 99
      read(nunit,3,err=99,end=99) nt,itxt
2     format(a)
      if(itxt.gt.0) call skip(itxt)
      errmsg='too much text'
      if(ideb.ne.1) write(6,97) errmsg
      if(nt.gt.ntitl) go to 99
      read(nunit,6,err=99,end=99) (text(i),i=1,nt)
6     format(80a1)
      do i=1,nt
       if(text(i).eq.char(0)) text(i)=' '
      end do 
c***** recopy first part into "short title"
      do i=1,20
       text(i+60)=text(i)
      end do
      indone=indone-1
      if(indone.eq.0) go to 20
      errmsg='31integer array'
      if(ideb.ne.1) write(6,97) errmsg
      read(nunit,2,err=99,end=99) rtype
      if(rtype.ne.'IIIIIIII') go to 99
      read(nunit,3,err=99,end=99) ni,itxt
      if(ni.ne.31) go to 99
      if(itxt.gt.0) call skip(itxt)
      read(nunit,3) (medpar(i),i=1,ni)
3     format(10i8)
      nvers=medpar(1)
c***** this is Allibon's format designator...
      errmsg=('data version')
      if(nvers.lt.3.or.nvers.gt.5) iflag=iflag+1
      if(iflag.gt.0.and.iflag.lt.3) Write(6,91) nvers
91    format(' Beware - data version ( ',i2,' ) mismatch ')
      indone=indone-1
      if(indone.eq.0) go to 20
      errmsg='real parameters'
      if(ideb.ne.1) write(6,97) errmsg
      read(nunit,2,err=99,end=99) rtype
      if(rtype.ne.'FFFFFFFF') go to 99
      read(nunit,3,err=99,end=99) np,itxt
      if(np.gt.npar) go to 99
      if(itxt.gt.0) call skip(itxt)
      read(nunit,7,err=99,end=99) (parr(i),i=1,np)
7     format(5e16.8)
      indone=indone-1
      if(indone.eq.0) go to 20
c***** next field should be SSSS
12    read(nunit,2,err=99,end=99) rtype
      if(rtype.ne.'SSSSSSSS') go to 12
      errmsg='spectrum header'
      if(ideb.ne.1) write(6,97) errmsg
      read(nunit,3,err=99,end=99) ispec,nrest,ntot,nrun,itxt
      if(itxt.gt.0) call skip(itxt)
      if(ntot.ne.1) go to 99
c***** should be one spectrum only
      errmsg='spectrum data'
      if(ideb.ne.1) write(6,97) errmsg
c***** now depending on date two possibilities (April '97)
      if(nvers.eq.3) go to 17 
c***** next - for version 4
      errmsg='real Sparameters'
      if(ideb.ne.1) write(6,97) errmsg
      read(nunit,2,err=99,end=99) rtype
      if(rtype.ne.'FFFFFFFF') go to 99
      read(nunit,3,err=99,end=99) nps,itxt
      if(nps.gt.10) go to 99
c***** only 5 at present ... but you never know
      if(itxt.gt.0) call skip(itxt)
      read(nunit,7,err=99,end=99) (parrs(i),i=1,nps)
c***** end of special for nvers=4      
17    read(nunit,2,err=99,end=99) rtype
      if(rtype.ne.'IIIIIIII') go to 99
      read(nunit,3,err=99,end=99) nd,itxt
      if(itxt.gt.0) call skip(itxt)
      errmsg='too many data'
      if(ideb.ne.1) write(6,97) errmsg
      if(nd.gt.ndx) go to 99

ccccccccccccccccccccccccccccccc sort out modus operandi for large data


      errmsg='counts data'
      read(nunit,3,err=99,end=99) (idon(i),i=1,nd)
20    close(nunit)
      call d16d22(ddate,text,ntitl,par,npar,idon,ndx,
     1medpar,ni,parr,nt,np,nd,parrs,nvers)
      ier=0
c***** sample detector distance (m)
      return
99    write(6,98) errmsg,fdev(1:lf)
98    format(' SASAS: Error - invalid - ',a,/
     1' Currently reading : ',a)
      close(nunit)
      return
      end
      subroutine d16d22(ddate,text,ntitl,par,npar,idon,ndx,
     1m,ni,parr,nt,np,nd,parrs,nvers)
c***** transcribes data to correspond to D22 definitions - March 1996
c
      real par(npar),parr(200),parrs(10)
      integer idon(ndx),nvers,m(ni)
      character*(*) ddate
      character*1 text(ntitl)
c***** text block - add date
      do 10 i=1,20
      text(100+i)=ddate(i:i)
10    continue
c***** real parameters
c
c***** preset 1
      par(1)=parr(39)
c***** time
      if(nvers.eq.3) par(3)=idon(3)/10.

c***** modified September 1999  clock=1khz

c***** in 0.1 secs
      if(nvers.ge.4) par(3)=parrs(1)/100.

c***** total detector counts
      if(nvers.eq.3) par(4)=idon(1)
      if(nvers.ge.4) par(4)=parrs(3)
c***** total monitor counts
      if(nvers.eq.3) par(5)=idon(2)
      if(nvers.ge.4) par(5)=parrs(2)
c***** dvm
      par(23)=parr(49)
c***** s-det distance mm in parr(30)
      par(26)=parr(30)/1000.
c***** temp set
      par(29)=parr(46)
c***** temp regul
      par(30)=parr(47)
c***** temp sample
      par(31)=parr(48)
c***** wavelength
      par(53)=parr(18)
c***** delta wave/wave
      par(54)=.02
c***** dx (mm per pixel)
      par(55)=2.54
c***** dy (mm per pixel)
      par(56)=5.0
c***** angles... twotheta and omega - either from header or scan data
c     twotheta(gamma)
      par(15)=parr(7)
c***** omega
      par(65)=parr(6)
c***** values at centre of scan .. check if scan variables
c     twotheta motor1, omega motor2
      do 1 i=1,7
c***** twotheta
      if(m(i+24).eq.1.and.nvers.eq.3) par(15)=idon(3+i)/100.
      if(m(i+24).eq.1.and.nvers.ge.4) par(15)=parrs(3+i)/1000.
c***** omega
      if(m(i+24).eq.2.and.nvers.eq.3) par(65)=(idon(3+i))/100.
      if(m(i+24).eq.2.and.nvers.ge.4) par(65)=(parrs(3+i))/1000.
1     continue
c***** now pack back integer data
      if(nvers.ge.4) go to 3
c***** mixed data no longer present - no backpacking
      do 2 i=1,1024
      idon(i)=idon(i+10)
2     continue
3     continue
      nt=120
      np=128
c      nd=1024

c***** new D16 detector - March 2002
      if(nd.eq.16384) then
c***** nominal detector centre - address
        par(51)=63.5
        par(52)=63.5
c***** dx (mm per pixel)
        par(55)=2.0
c***** dy (mm per pixel)
        par(56)=2.0
        call delege(idon)
      end if
c***** new MILAND 32cm x 32cm 1mm resolution 
      if(nd.eq.102400) then
c***** nominal detector centre - address
        par(51)=159.5
        par(52)=159.5
c***** dx (mm per pixel)
        par(55)=1.0
c***** dy (mm per pixel)
        par(56)=1.0
c        call delege1(idon)
         call transxy(idon,320)
      end if

      return
      end
      subroutine skip(nskip)
      common/sasasc/nunit
      do 1 i=1,nskip
      read(nunit,2)
2     format(1x)
1     continue
      return
      end
      subroutine delege(idon)
c***** D16 delete first row and first column of 16K detector data
      dimension idon(128,128)
      do i=1,128
       idon(1,i)=0
       idon(i,1)=0
      end do
      return
      end
      subroutine delege1(idon)
c***** D16 delete first row and first column of 1M detector data
      dimension idon(320,320),itmp(320,320)
c***** invert x and y
      do i=1,320
      do j=1,320
      itmp(i,j)=idon(j,i)
      end do
      end do
      do i=1,320
      do j=1,320
      idon(i,j)=itmp(i,j)
      end do
      end do
      do i=1,320
       idon(1,i)=0
       idon(i,1)=0
      end do
      return
      end
      subroutine transxy(in,nd)
      dimension in(nd,nd)
c***** move rows to columns
      do iy=1,nd
      do ix=1,iy
      itmp=in(ix,iy)
      itmpp=in(iy,ix)
      in(ix,iy)=itmpp
      in(iy,ix)=itmp
      end do
      end do
      return
      end
      subroutine cropen(fdev,lf,nunit,ier)
c***** looks for numor, numor.Z or numor.gz (May 2008)
      common/sdbg/ideb
      logical found
      character*(*) fdev
      character*140 cmd
      character*100 infpro
      character*20 cached,cachee
      character*16 errmsg
      cached=' '
      cachee=' '
      call getenv('SAS_CACHE',cachee)
      call trimmd(cachee,lc)
      if(lc.gt.13) lc=0
      ier=-1
c***** first look for local cached file - need only run number - last 6
      if(lc.eq.0) cached='cashas/'//fdev(lf-5:lf)
      if(lc.gt.0) cached=cachee(1:lc)//'/'//fdev(lf-5:lf)
c***** default (temporary) is cashas (xplot)
      errmsg='cache seek'
      if(ideb.ne.1) write(6,97) errmsg
 97    format(' SASAS - debug reached test for ',a)
      if(ideb.ne.1) write(6,91) cached
 91   format(' SASAS - seeking ',a)
      inquire(file=cached,exist=found)
      if(.not.found) go to 5
      open(unit=nunit,file=cached,status='old',err=99)
      ier=0
      return
5     errmsg='raw file'
      if(ideb.ne.1) write(6,97) errmsg
      if(ideb.ne.1) write(6,91) fdev(1:lf)
      inquire(file=fdev(1:lf),exist=found)
c***** if found then open file
      if(.not.found) go to 20
      open(unit=nunit,file=fdev,status='old',err=99)
      ier=0
      return
20    continue
c***** try compressed file..
      call getenv('SAS_DECOMPRESS',infpro)
      if(index(infpro,' ').eq.1) return
c***** no decompress program available
      call trimmd(infpro,linfp)
      errmsg='compressed file'
      if(ideb.ne.1) write(6,97) errmsg
      fdev(lf+1:lf+2)='.Z'
      lt=lf+2
      if(ideb.ne.1) write(6,91) fdev(1:lt)
      inquire(file=fdev(1:lt),exist=found)
c      write(*,*)' fdev - found ',fdev(1:lf),found 
      if(found) go to 30
      fdev(lf+1:lf+3)='.gz'
      lt=lf+3
      if(ideb.ne.1) write(6,91) fdev(1:lt)
      inquire(file=fdev(1:lt),exist=found)
      if(found) go to 30
c***** lets not complicate filenames with .Z in message
      return
30    cmd=infpro(1:linfp+1)//fdev(1:lf)//' >sasas.tmp'
      call system(cmd)
      open(unit=nunit,file='sasas.tmp',status='old',err=99)
      ier=0
99    return
      end
      block data sasasb
      common/maxarr/ntitl,npar,nparx,ndx
      common/sasasc/nunit
      common/liminp/ininp
      COMMON/NEWMASK/GRDD
      save
      character*8 GRDD
      data GRDD/'UNMASKED'/
c***** SASAS is always present when masking is employed
      data ntitl/512/,npar/128/,nparx/20/,ndx/16384/
      data nunit/31/
      data ininp/-1/
      end
